import { Injectable, defineInjectable } from '@angular/core';
import 'reflect-metadata';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class MappingOptions {
    /**
     * @param {?} className
     * @param {?} classPropertyName
     * @param {?} jsonPropertyName
     * @param {?=} isArray
     * @param {?=} expectedJsonType
     * @param {?=} isOptional
     * @param {?=} customMapper
     * @param {?=} notNull
     * @param {?=} overrideInitValue
     */
    constructor(className, classPropertyName, jsonPropertyName, isArray = false, expectedJsonType = null, isOptional = false, customMapper = null, notNull = false, overrideInitValue = false) {
        this.className = className;
        this.classPropertyName = classPropertyName;
        this.jsonPropertyName = jsonPropertyName;
        this.isArray = isArray;
        this.expectedJsonType = expectedJsonType;
        this.isOptional = isOptional;
        this.customMapper = customMapper;
        this.notNull = notNull;
        this.overrideInitValue = overrideInitValue;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const JSON_OBJECT = 'jtsm_json_object';
/** @type {?} */
const CUSTOM_CONVERTER = 'jtsm_custom_converter';
/** @type {?} */
const MAPPING_OPTIONS = 'jtsm_mapping_options';
/** @type {?} */
const MAPPING_OPTIONS_OPTIONAL = 'jtsm_mapping_optional';
/** @type {?} */
const MAPPING_OPTIONS_NOTNULL = 'jtsm_mapping_not_null';
/** @type {?} */
const MAPPING_OPTIONS_MAPPER = 'jtsm_mapping_mapper';
/**
 * @param {?} target
 * @return {?}
 */
function JsonConverter(target) {
    Reflect.defineMetadata(CUSTOM_CONVERTER, '__json_converter__', target);
}
/**
 * @param {?} target
 * @return {?}
 */
function JsonObject(target) {
    Reflect.defineMetadata(JSON_OBJECT, '__json_object__', target);
    Reflect.getMetadataKeys(target.prototype)
        .filter((/**
     * @param {?} key
     * @return {?}
     */
    key => {
        return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
    }))
        .forEach((/**
     * @param {?} key
     * @return {?}
     */
    key => {
        Reflect.defineMetadata(key, Reflect.getMetadata(key, target.prototype), target);
    }));
}
/**
 * @param {?} options
 * @return {?}
 */
function JsonObjectOptions(options) {
    return (/**
     * @param {?} target
     * @return {?}
     */
    function (target) {
        Reflect.defineMetadata(JSON_OBJECT, '__json_object__', target);
        Reflect.getMetadataKeys(target.prototype)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        key => {
            return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
        }))
            .forEach((/**
         * @param {?} key
         * @return {?}
         */
        (key) => {
            /** @type {?} */
            const metadata = Reflect.getMetadata(key, target.prototype);
            if (key.startsWith(MAPPING_OPTIONS_MAPPER) && options.overrideInitValues !== undefined) {
                ((/** @type {?} */ (metadata))).overrideInitValue = options.overrideInitValues;
            }
            Reflect.defineMetadata(key, metadata, target);
        }));
    });
}
/**
 * @param {?} jsonPropertyName
 * @param {?} expectedType
 * @return {?}
 */
function JsonProperty(jsonPropertyName, expectedType) {
    return (/**
     * @param {?} target
     * @param {?} classPropertyName
     * @return {?}
     */
    function (target, classPropertyName) {
        /** @type {?} */
        let isArray = false;
        if (Array.isArray(expectedType)) {
            if (expectedType.length !== 1) {
                throw new Error('JsonTsMapper : @JsonProperty : Type array must have length = 1');
            }
            isArray = true;
            expectedType = expectedType[0];
        }
        jsonPropertyName = jsonPropertyName === undefined ? classPropertyName : jsonPropertyName;
        /** @type {?} */
        const options = new MappingOptions(target.constructor.name, classPropertyName, jsonPropertyName, isArray, expectedType);
        if (Reflect.hasMetadata(`${MAPPING_OPTIONS_OPTIONAL}-${classPropertyName}`, target)) {
            options.isOptional = true; // Reflect.getMetadata(`${MAPPING_OPTIONS_OPTIONAL}-${classPropertyName}`, target);
        }
        if (Reflect.hasMetadata(`${MAPPING_OPTIONS_NOTNULL}-${classPropertyName}`, target)) {
            options.notNull = true; // Reflect.getMetadata(`${MAPPING_OPTIONS_NOTNULL}-${classPropertyName}`, target);
            options.isOptional = false;
        }
        if (Reflect.hasMetadata(`${MAPPING_OPTIONS_MAPPER}-${classPropertyName}`, target)) {
            options.customMapper = Reflect.getMetadata(`${MAPPING_OPTIONS_MAPPER}-${classPropertyName}`, target);
        }
        Reflect.defineMetadata(`${MAPPING_OPTIONS}-${classPropertyName}`, options, target);
    });
}
/**
 * @param {?} target
 * @param {?} classPropertyName
 * @return {?}
 */
function Optional(target, classPropertyName) {
    if (Reflect.hasMetadata(`${MAPPING_OPTIONS_NOTNULL}-${classPropertyName}`, target)) {
        throw new Error('"Optional" and "NotNull" annotations are not compatible');
    }
    Reflect.defineMetadata(`${MAPPING_OPTIONS_OPTIONAL}-${classPropertyName}`, true, target);
}
/**
 * @param {?} target
 * @param {?} classPropertyName
 * @return {?}
 */
function NotNull(target, classPropertyName) {
    if (Reflect.hasMetadata(`${MAPPING_OPTIONS_OPTIONAL}-${classPropertyName}`, target)) {
        throw new Error('"@Optional" and "@NotNull" annotations are not compatible');
    }
    Reflect.defineMetadata(`${MAPPING_OPTIONS_NOTNULL}-${classPropertyName}`, true, target);
}
/**
 * @template T
 * @param {?} customMapper
 * @return {?}
 */
function CustomConverter(customMapper) {
    return (/**
     * @param {?} target
     * @param {?} classPropertyName
     * @return {?}
     */
    function (target, classPropertyName) {
        Reflect.defineMetadata(`${MAPPING_OPTIONS_MAPPER}-${classPropertyName}`, customMapper, target);
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class Any {
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class JsonTsMapperService {
    /**
     * Serialize objct to 'json object'
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?} json object
     */
    serialize(instance, context) {
        if (Array.isArray(instance)) {
            return instance.map((/**
             * @param {?} item
             * @return {?}
             */
            item => this.serialize_loopProperties(item, context)));
        }
        else {
            return this.serialize_loopProperties(instance, context);
        }
    }
    /**
     * Serialize objct to json string
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?} json string
     */
    serializeToString(instance, context) {
        return JSON.stringify(this.serialize(instance, context));
    }
    /**
     * Deserialize an instance or array of {T}
     * @template T
     * @param {?} obj object, array or json string to deserialize
     * @param {?} clazz class to map object
     * @param {?=} context
     * @return {?} instance or array of {T}
     */
    deserialize(obj, clazz, context) {
        if (typeof obj === 'string') {
            obj = JSON.parse(obj);
        }
        if (Array.isArray(obj)) {
            return this.deserializeArray(obj, clazz, context);
        }
        else {
            return this.deserializeObject(obj, clazz, context);
        }
    }
    /**
     * Deserialize {T}
     * @template T
     * @param {?} obj object or json string to deserialize
     * @param {?} clazz class to map object
     * @param {?=} context
     * @return {?} instance of {T}
     */
    deserializeObject(obj, clazz, context) {
        if (typeof obj === 'string') {
            obj = JSON.parse(obj);
        }
        obj = (/** @type {?} */ (obj));
        return this.deserialize_loopProperties(obj, clazz, context);
    }
    /**
     * Deserialize array of {T}
     * @template T
     * @param {?} obj array or json string to deserialize
     * @param {?} clazz class to map objects
     * @param {?=} context
     * @return {?} T[]
     */
    deserializeArray(obj, clazz, context) {
        if (typeof obj === 'string') {
            obj = JSON.parse(obj);
        }
        /** @type {?} */
        const list = (/** @type {?} */ (obj));
        return list.map((/**
         * @param {?} item
         * @return {?}
         */
        item => this.deserializeObject(item, clazz, context)));
    }
    /**
     * @private
     * @template T
     * @param {?} obj
     * @param {?} clazz
     * @param {?=} context
     * @return {?}
     */
    deserialize_loopProperties(obj, clazz, context) {
        if (Reflect.hasMetadata(JSON_OBJECT, clazz)) {
            /** @type {?} */
            let instance = new clazz();
            /** @type {?} */
            const mappingOptionsKeys = Reflect.getMetadataKeys(clazz)
                .filter((/**
             * @param {?} key
             * @return {?}
             */
            key => {
                return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
            }));
            mappingOptionsKeys.forEach((/**
             * @param {?} key
             * @return {?}
             */
            key => {
                /** @type {?} */
                const options = Reflect.getMetadata(key, clazz);
                /** @type {?} */
                const instanceInitValue = instance[options.classPropertyName];
                if (options.isArray) {
                    instance[options.classPropertyName] = this.deserializePropertyArray(obj[options.jsonPropertyName], options, context);
                }
                else {
                    instance[options.classPropertyName] = this.deserializeProperty(obj[options.jsonPropertyName], options, context);
                }
                if (instance[options.classPropertyName] === undefined && options.overrideInitValue === false) {
                    instance[options.classPropertyName] = instanceInitValue;
                }
            }));
            return instance;
        }
        else {
            throw new Error(`JsonTsMapper | class : '${clazz.name}' | Trying to deserialize an object wihout @JsonObject annotation`);
        }
    }
    /**
     * @private
     * @param {?} jsonValues
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    deserializePropertyArray(jsonValues, options, context) {
        if (jsonValues === undefined || jsonValues === null) {
            if (options.isOptional === false && jsonValues === undefined) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | deserialize : json property ${options.jsonPropertyName} is missing and not optional`);
            }
            else if (options.notNull === true && jsonValues === null) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | deserialize : json property ${options.jsonPropertyName} is null and set as 'notNull'`);
            }
            else {
                return jsonValues;
            }
        }
        if (!this.checkType(jsonValues, options.expectedJsonType, true)) {
            throw new Error(`JsonTsMapper | class : '${options.className}' | serialize : type missmatch for property ${options.jsonPropertyName} : expected type : Array, received type : ${typeof jsonValues}`);
        }
        return jsonValues.map((/**
         * @param {?} jsonValue
         * @return {?}
         */
        jsonValue => this.deserializeProperty(jsonValue, options, context)));
    }
    /**
     * @private
     * @param {?} jsonValue
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    deserializeProperty(jsonValue, options, context) {
        if (jsonValue === undefined || jsonValue === null) {
            if (options.isOptional === false && jsonValue === undefined) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | deserialize : json property ${options.jsonPropertyName} is missing and not optional`);
            }
            else if (options.notNull === true && jsonValue === null) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | deserialize : json property ${options.jsonPropertyName} is null and set as 'notNull'`);
            }
            else {
                return jsonValue;
            }
        }
        if (!this.checkType(jsonValue, options.expectedJsonType, false)) {
            throw new Error(`JsonTsMapper | class : '${options.className}' | deserialize : type missmatch for property ${options.jsonPropertyName} : expected type : ${options.expectedJsonType.name}, received type : ${typeof jsonValue}`);
        }
        if (options.customMapper) {
            /** @type {?} */
            const mapper = new options.customMapper();
            return mapper.deserialize(jsonValue, context);
        }
        else if (Reflect.hasMetadata(JSON_OBJECT, options.expectedJsonType)) {
            return this.deserialize(jsonValue, options.expectedJsonType, context);
        }
        else {
            return jsonValue;
        }
    }
    /**
     * @private
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?}
     */
    serialize_loopProperties(instance, context) {
        if (Reflect.hasMetadata(JSON_OBJECT, instance.constructor)) {
            /** @type {?} */
            let json = {};
            /** @type {?} */
            const mappingOptionsKeys = Reflect.getMetadataKeys(instance.constructor)
                .filter((/**
             * @param {?} key
             * @return {?}
             */
            key => {
                return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
            }));
            mappingOptionsKeys.forEach((/**
             * @param {?} key
             * @return {?}
             */
            key => {
                /** @type {?} */
                const options = Reflect.getMetadata(key, instance.constructor);
                if (options.isArray) {
                    json[options.jsonPropertyName] = this.serializePropertyArray(instance[options.classPropertyName], options, context);
                }
                else {
                    json[options.jsonPropertyName] = this.serializeProperty(instance[options.classPropertyName], options, context);
                }
            }));
            for (let key in json) {
                if (json[key] === undefined) {
                    delete json[key];
                }
            }
            return json;
        }
        else {
            throw new Error(`JsonTsMapper : Trying to serialize an object wihout @JsonObject annotation`);
        }
    }
    /**
     * @private
     * @param {?} values
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    serializePropertyArray(values, options, context) {
        if (values === null || values === undefined) {
            if (options.isOptional === false && values === undefined) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | serialize : class property ${options.classPropertyName} is missing and not optional`);
            }
            else if (options.notNull === true && values === null) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | serialize : class property ${options.classPropertyName} is null and set as 'notNull'`);
            }
            else {
                return values;
            }
        }
        if (!Array.isArray(values)) {
            throw new Error(`JsonTsMapper | class : '${options.className}' | serialize : type missmatch for property ${options.classPropertyName} : expected type : Array, serialized type : ${typeof values}`);
        }
        return values.map((/**
         * @param {?} value
         * @return {?}
         */
        (value) => this.serializeProperty(value, options, context)));
    }
    /**
     * @private
     * @param {?} value
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    serializeProperty(value, options, context) {
        /** @type {?} */
        let jsonValue;
        if (value === undefined || value === null) {
            jsonValue = value;
        }
        else {
            if (options.customMapper) {
                /** @type {?} */
                const mapper = new options.customMapper();
                jsonValue = mapper.serialize(value, context);
            }
            else if (Reflect.hasMetadata(JSON_OBJECT, options.expectedJsonType)) {
                jsonValue = this.serialize(value, context);
            }
            else {
                jsonValue = value;
            }
        }
        if (jsonValue === null || jsonValue === undefined) {
            if (options.isOptional === false && jsonValue === undefined) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | serialize : class property ${options.classPropertyName} is missing and not optional`);
            }
            else if (options.notNull === true && jsonValue === null) {
                throw new Error(`JsonTsMapper | class : '${options.className}' | serialize : class property ${options.classPropertyName} is null and set as 'notNull'`);
            }
            else {
                return jsonValue;
            }
        }
        if (!this.checkType(jsonValue, options.expectedJsonType, false)) {
            throw new Error(`JsonTsMapper | class : '${options.className}' | serialize : type missmatch for property ${options.classPropertyName}: expected type : ${options.expectedJsonType.name}, serialized type : ${typeof jsonValue}`);
        }
        return jsonValue;
    }
    /**
     * @param {?} jsonValue
     * @param {?=} expectedJsonType
     * @param {?=} isArray
     * @return {?}
     */
    checkType(jsonValue, expectedJsonType = Any, isArray = false) {
        if (isArray) {
            if (Array.isArray(jsonValue)) {
                if (jsonValue.length) {
                    jsonValue = jsonValue[0];
                }
                else {
                    return true;
                }
            }
            else {
                return false;
            }
        }
        switch (expectedJsonType) {
            case Any:
                return true;
            case String:
                return typeof jsonValue === 'string';
            case Number:
                return typeof jsonValue === 'number';
            case Boolean:
                return typeof jsonValue === 'boolean';
            default:
                return typeof jsonValue === 'object';
        }
    }
}
JsonTsMapperService.decorators = [
    { type: Injectable, args: [{
                providedIn: 'root'
            },] }
];
/** @nocollapse */ JsonTsMapperService.ngInjectableDef = defineInjectable({ factory: function JsonTsMapperService_Factory() { return new JsonTsMapperService(); }, token: JsonTsMapperService, providedIn: "root" });

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @abstract
 * @template JsonType, TsType
 */
class AbstractJsonConverter {
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { JsonObject, JsonObjectOptions, JsonProperty, Optional, NotNull, CustomConverter, JsonConverter, Any, JsonTsMapperService, AbstractJsonConverter };

//# sourceMappingURL=thorolf-json-ts-mapper.js.map