import { Injectable, defineInjectable } from '@angular/core';
import 'reflect-metadata';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var MappingOptions = /** @class */ (function () {
    function MappingOptions(className, classPropertyName, jsonPropertyName, isArray, expectedJsonType, isOptional, customMapper, notNull, overrideInitValue) {
        if (isArray === void 0) { isArray = false; }
        if (expectedJsonType === void 0) { expectedJsonType = null; }
        if (isOptional === void 0) { isOptional = false; }
        if (customMapper === void 0) { customMapper = null; }
        if (notNull === void 0) { notNull = false; }
        if (overrideInitValue === void 0) { overrideInitValue = false; }
        this.className = className;
        this.classPropertyName = classPropertyName;
        this.jsonPropertyName = jsonPropertyName;
        this.isArray = isArray;
        this.expectedJsonType = expectedJsonType;
        this.isOptional = isOptional;
        this.customMapper = customMapper;
        this.notNull = notNull;
        this.overrideInitValue = overrideInitValue;
    }
    return MappingOptions;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var JSON_OBJECT = 'jtsm_json_object';
/** @type {?} */
var CUSTOM_CONVERTER = 'jtsm_custom_converter';
/** @type {?} */
var MAPPING_OPTIONS = 'jtsm_mapping_options';
/** @type {?} */
var MAPPING_OPTIONS_OPTIONAL = 'jtsm_mapping_optional';
/** @type {?} */
var MAPPING_OPTIONS_NOTNULL = 'jtsm_mapping_not_null';
/** @type {?} */
var MAPPING_OPTIONS_MAPPER = 'jtsm_mapping_mapper';
/**
 * @param {?} target
 * @return {?}
 */
function JsonConverter(target) {
    Reflect.defineMetadata(CUSTOM_CONVERTER, '__json_converter__', target);
}
/**
 * @param {?} target
 * @return {?}
 */
function JsonObject(target) {
    Reflect.defineMetadata(JSON_OBJECT, '__json_object__', target);
    Reflect.getMetadataKeys(target.prototype)
        .filter((/**
     * @param {?} key
     * @return {?}
     */
    function (key) {
        return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
    }))
        .forEach((/**
     * @param {?} key
     * @return {?}
     */
    function (key) {
        Reflect.defineMetadata(key, Reflect.getMetadata(key, target.prototype), target);
    }));
}
/**
 * @param {?} options
 * @return {?}
 */
function JsonObjectOptions(options) {
    return (/**
     * @param {?} target
     * @return {?}
     */
    function (target) {
        Reflect.defineMetadata(JSON_OBJECT, '__json_object__', target);
        Reflect.getMetadataKeys(target.prototype)
            .filter((/**
         * @param {?} key
         * @return {?}
         */
        function (key) {
            return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
        }))
            .forEach((/**
         * @param {?} key
         * @return {?}
         */
        function (key) {
            /** @type {?} */
            var metadata = Reflect.getMetadata(key, target.prototype);
            if (key.startsWith(MAPPING_OPTIONS_MAPPER) && options.overrideInitValues !== undefined) {
                ((/** @type {?} */ (metadata))).overrideInitValue = options.overrideInitValues;
            }
            Reflect.defineMetadata(key, metadata, target);
        }));
    });
}
/**
 * @param {?} jsonPropertyName
 * @param {?} expectedType
 * @return {?}
 */
function JsonProperty(jsonPropertyName, expectedType) {
    return (/**
     * @param {?} target
     * @param {?} classPropertyName
     * @return {?}
     */
    function (target, classPropertyName) {
        /** @type {?} */
        var isArray = false;
        if (Array.isArray(expectedType)) {
            if (expectedType.length !== 1) {
                throw new Error('JsonTsMapper : @JsonProperty : Type array must have length = 1');
            }
            isArray = true;
            expectedType = expectedType[0];
        }
        jsonPropertyName = jsonPropertyName === undefined ? classPropertyName : jsonPropertyName;
        /** @type {?} */
        var options = new MappingOptions(target.constructor.name, classPropertyName, jsonPropertyName, isArray, expectedType);
        if (Reflect.hasMetadata(MAPPING_OPTIONS_OPTIONAL + "-" + classPropertyName, target)) {
            options.isOptional = true; // Reflect.getMetadata(`${MAPPING_OPTIONS_OPTIONAL}-${classPropertyName}`, target);
        }
        if (Reflect.hasMetadata(MAPPING_OPTIONS_NOTNULL + "-" + classPropertyName, target)) {
            options.notNull = true; // Reflect.getMetadata(`${MAPPING_OPTIONS_NOTNULL}-${classPropertyName}`, target);
            options.isOptional = false;
        }
        if (Reflect.hasMetadata(MAPPING_OPTIONS_MAPPER + "-" + classPropertyName, target)) {
            options.customMapper = Reflect.getMetadata(MAPPING_OPTIONS_MAPPER + "-" + classPropertyName, target);
        }
        Reflect.defineMetadata(MAPPING_OPTIONS + "-" + classPropertyName, options, target);
    });
}
/**
 * @param {?} target
 * @param {?} classPropertyName
 * @return {?}
 */
function Optional(target, classPropertyName) {
    if (Reflect.hasMetadata(MAPPING_OPTIONS_NOTNULL + "-" + classPropertyName, target)) {
        throw new Error('"Optional" and "NotNull" annotations are not compatible');
    }
    Reflect.defineMetadata(MAPPING_OPTIONS_OPTIONAL + "-" + classPropertyName, true, target);
}
/**
 * @param {?} target
 * @param {?} classPropertyName
 * @return {?}
 */
function NotNull(target, classPropertyName) {
    if (Reflect.hasMetadata(MAPPING_OPTIONS_OPTIONAL + "-" + classPropertyName, target)) {
        throw new Error('"@Optional" and "@NotNull" annotations are not compatible');
    }
    Reflect.defineMetadata(MAPPING_OPTIONS_NOTNULL + "-" + classPropertyName, true, target);
}
/**
 * @template T
 * @param {?} customMapper
 * @return {?}
 */
function CustomConverter(customMapper) {
    return (/**
     * @param {?} target
     * @param {?} classPropertyName
     * @return {?}
     */
    function (target, classPropertyName) {
        Reflect.defineMetadata(MAPPING_OPTIONS_MAPPER + "-" + classPropertyName, customMapper, target);
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var Any = /** @class */ (function () {
    function Any() {
    }
    return Any;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
var JsonTsMapperService = /** @class */ (function () {
    function JsonTsMapperService() {
    }
    /**
     * Serialize objct to 'json object'
     * @param  obj object or array to serialize
     * @return    json object
     */
    /**
     * Serialize objct to 'json object'
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?} json object
     */
    JsonTsMapperService.prototype.serialize = /**
     * Serialize objct to 'json object'
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?} json object
     */
    function (instance, context) {
        var _this = this;
        if (Array.isArray(instance)) {
            return instance.map((/**
             * @param {?} item
             * @return {?}
             */
            function (item) { return _this.serialize_loopProperties(item, context); }));
        }
        else {
            return this.serialize_loopProperties(instance, context);
        }
    };
    /**
     * Serialize objct to json string
     * @param  obj object or array to serialize
     * @return    json string
     */
    /**
     * Serialize objct to json string
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?} json string
     */
    JsonTsMapperService.prototype.serializeToString = /**
     * Serialize objct to json string
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?} json string
     */
    function (instance, context) {
        return JSON.stringify(this.serialize(instance, context));
    };
    /**
     * Deserialize an instance or array of {T}
     * @param obj object, array or json string to deserialize
     * @param clazz class to map object
     * @return instance or array of {T}
     */
    /**
     * Deserialize an instance or array of {T}
     * @template T
     * @param {?} obj object, array or json string to deserialize
     * @param {?} clazz class to map object
     * @param {?=} context
     * @return {?} instance or array of {T}
     */
    JsonTsMapperService.prototype.deserialize = /**
     * Deserialize an instance or array of {T}
     * @template T
     * @param {?} obj object, array or json string to deserialize
     * @param {?} clazz class to map object
     * @param {?=} context
     * @return {?} instance or array of {T}
     */
    function (obj, clazz, context) {
        if (typeof obj === 'string') {
            obj = JSON.parse(obj);
        }
        if (Array.isArray(obj)) {
            return this.deserializeArray(obj, clazz, context);
        }
        else {
            return this.deserializeObject(obj, clazz, context);
        }
    };
    /**
     * Deserialize {T}
     * @param obj object or json string to deserialize
     * @param clazz class to map object
     * @return instance of {T}
     */
    /**
     * Deserialize {T}
     * @template T
     * @param {?} obj object or json string to deserialize
     * @param {?} clazz class to map object
     * @param {?=} context
     * @return {?} instance of {T}
     */
    JsonTsMapperService.prototype.deserializeObject = /**
     * Deserialize {T}
     * @template T
     * @param {?} obj object or json string to deserialize
     * @param {?} clazz class to map object
     * @param {?=} context
     * @return {?} instance of {T}
     */
    function (obj, clazz, context) {
        if (typeof obj === 'string') {
            obj = JSON.parse(obj);
        }
        obj = (/** @type {?} */ (obj));
        return this.deserialize_loopProperties(obj, clazz, context);
    };
    /**
     * Deserialize array of {T}
     * @param obj array or json string to deserialize
     * @param clazz class to map objects
     * @return T[]
     */
    /**
     * Deserialize array of {T}
     * @template T
     * @param {?} obj array or json string to deserialize
     * @param {?} clazz class to map objects
     * @param {?=} context
     * @return {?} T[]
     */
    JsonTsMapperService.prototype.deserializeArray = /**
     * Deserialize array of {T}
     * @template T
     * @param {?} obj array or json string to deserialize
     * @param {?} clazz class to map objects
     * @param {?=} context
     * @return {?} T[]
     */
    function (obj, clazz, context) {
        var _this = this;
        if (typeof obj === 'string') {
            obj = JSON.parse(obj);
        }
        /** @type {?} */
        var list = (/** @type {?} */ (obj));
        return list.map((/**
         * @param {?} item
         * @return {?}
         */
        function (item) { return _this.deserializeObject(item, clazz, context); }));
    };
    /**
     * @private
     * @template T
     * @param {?} obj
     * @param {?} clazz
     * @param {?=} context
     * @return {?}
     */
    JsonTsMapperService.prototype.deserialize_loopProperties = /**
     * @private
     * @template T
     * @param {?} obj
     * @param {?} clazz
     * @param {?=} context
     * @return {?}
     */
    function (obj, clazz, context) {
        var _this = this;
        if (Reflect.hasMetadata(JSON_OBJECT, clazz)) {
            /** @type {?} */
            var instance_1 = new clazz();
            /** @type {?} */
            var mappingOptionsKeys = Reflect.getMetadataKeys(clazz)
                .filter((/**
             * @param {?} key
             * @return {?}
             */
            function (key) {
                return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
            }));
            mappingOptionsKeys.forEach((/**
             * @param {?} key
             * @return {?}
             */
            function (key) {
                /** @type {?} */
                var options = Reflect.getMetadata(key, clazz);
                /** @type {?} */
                var instanceInitValue = instance_1[options.classPropertyName];
                if (options.isArray) {
                    instance_1[options.classPropertyName] = _this.deserializePropertyArray(obj[options.jsonPropertyName], options, context);
                }
                else {
                    instance_1[options.classPropertyName] = _this.deserializeProperty(obj[options.jsonPropertyName], options, context);
                }
                if (instance_1[options.classPropertyName] === undefined && options.overrideInitValue === false) {
                    instance_1[options.classPropertyName] = instanceInitValue;
                }
            }));
            return instance_1;
        }
        else {
            throw new Error("JsonTsMapper | class : '" + clazz.name + "' | Trying to deserialize an object wihout @JsonObject annotation");
        }
    };
    /**
     * @private
     * @param {?} jsonValues
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    JsonTsMapperService.prototype.deserializePropertyArray = /**
     * @private
     * @param {?} jsonValues
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    function (jsonValues, options, context) {
        var _this = this;
        if (jsonValues === undefined || jsonValues === null) {
            if (options.isOptional === false && jsonValues === undefined) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | deserialize : json property " + options.jsonPropertyName + " is missing and not optional");
            }
            else if (options.notNull === true && jsonValues === null) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | deserialize : json property " + options.jsonPropertyName + " is null and set as 'notNull'");
            }
            else {
                return jsonValues;
            }
        }
        if (!this.checkType(jsonValues, options.expectedJsonType, true)) {
            throw new Error("JsonTsMapper | class : '" + options.className + "' | serialize : type missmatch for property " + options.jsonPropertyName + " : expected type : Array, received type : " + typeof jsonValues);
        }
        return jsonValues.map((/**
         * @param {?} jsonValue
         * @return {?}
         */
        function (jsonValue) { return _this.deserializeProperty(jsonValue, options, context); }));
    };
    /**
     * @private
     * @param {?} jsonValue
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    JsonTsMapperService.prototype.deserializeProperty = /**
     * @private
     * @param {?} jsonValue
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    function (jsonValue, options, context) {
        if (jsonValue === undefined || jsonValue === null) {
            if (options.isOptional === false && jsonValue === undefined) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | deserialize : json property " + options.jsonPropertyName + " is missing and not optional");
            }
            else if (options.notNull === true && jsonValue === null) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | deserialize : json property " + options.jsonPropertyName + " is null and set as 'notNull'");
            }
            else {
                return jsonValue;
            }
        }
        if (!this.checkType(jsonValue, options.expectedJsonType, false)) {
            throw new Error("JsonTsMapper | class : '" + options.className + "' | deserialize : type missmatch for property " + options.jsonPropertyName + " : expected type : " + options.expectedJsonType.name + ", received type : " + typeof jsonValue);
        }
        if (options.customMapper) {
            /** @type {?} */
            var mapper = new options.customMapper();
            return mapper.deserialize(jsonValue, context);
        }
        else if (Reflect.hasMetadata(JSON_OBJECT, options.expectedJsonType)) {
            return this.deserialize(jsonValue, options.expectedJsonType, context);
        }
        else {
            return jsonValue;
        }
    };
    /**
     * @private
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?}
     */
    JsonTsMapperService.prototype.serialize_loopProperties = /**
     * @private
     * @template T
     * @param {?} instance
     * @param {?=} context
     * @return {?}
     */
    function (instance, context) {
        var _this = this;
        if (Reflect.hasMetadata(JSON_OBJECT, instance.constructor)) {
            /** @type {?} */
            var json_1 = {};
            /** @type {?} */
            var mappingOptionsKeys = Reflect.getMetadataKeys(instance.constructor)
                .filter((/**
             * @param {?} key
             * @return {?}
             */
            function (key) {
                return typeof key === 'string' && key.startsWith(MAPPING_OPTIONS);
            }));
            mappingOptionsKeys.forEach((/**
             * @param {?} key
             * @return {?}
             */
            function (key) {
                /** @type {?} */
                var options = Reflect.getMetadata(key, instance.constructor);
                if (options.isArray) {
                    json_1[options.jsonPropertyName] = _this.serializePropertyArray(instance[options.classPropertyName], options, context);
                }
                else {
                    json_1[options.jsonPropertyName] = _this.serializeProperty(instance[options.classPropertyName], options, context);
                }
            }));
            for (var key in json_1) {
                if (json_1[key] === undefined) {
                    delete json_1[key];
                }
            }
            return json_1;
        }
        else {
            throw new Error("JsonTsMapper : Trying to serialize an object wihout @JsonObject annotation");
        }
    };
    /**
     * @private
     * @param {?} values
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    JsonTsMapperService.prototype.serializePropertyArray = /**
     * @private
     * @param {?} values
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    function (values, options, context) {
        var _this = this;
        if (values === null || values === undefined) {
            if (options.isOptional === false && values === undefined) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | serialize : class property " + options.classPropertyName + " is missing and not optional");
            }
            else if (options.notNull === true && values === null) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | serialize : class property " + options.classPropertyName + " is null and set as 'notNull'");
            }
            else {
                return values;
            }
        }
        if (!Array.isArray(values)) {
            throw new Error("JsonTsMapper | class : '" + options.className + "' | serialize : type missmatch for property " + options.classPropertyName + " : expected type : Array, serialized type : " + typeof values);
        }
        return values.map((/**
         * @param {?} value
         * @return {?}
         */
        function (value) { return _this.serializeProperty(value, options, context); }));
    };
    /**
     * @private
     * @param {?} value
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    JsonTsMapperService.prototype.serializeProperty = /**
     * @private
     * @param {?} value
     * @param {?} options
     * @param {?=} context
     * @return {?}
     */
    function (value, options, context) {
        /** @type {?} */
        var jsonValue;
        if (value === undefined || value === null) {
            jsonValue = value;
        }
        else {
            if (options.customMapper) {
                /** @type {?} */
                var mapper = new options.customMapper();
                jsonValue = mapper.serialize(value, context);
            }
            else if (Reflect.hasMetadata(JSON_OBJECT, options.expectedJsonType)) {
                jsonValue = this.serialize(value, context);
            }
            else {
                jsonValue = value;
            }
        }
        if (jsonValue === null || jsonValue === undefined) {
            if (options.isOptional === false && jsonValue === undefined) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | serialize : class property " + options.classPropertyName + " is missing and not optional");
            }
            else if (options.notNull === true && jsonValue === null) {
                throw new Error("JsonTsMapper | class : '" + options.className + "' | serialize : class property " + options.classPropertyName + " is null and set as 'notNull'");
            }
            else {
                return jsonValue;
            }
        }
        if (!this.checkType(jsonValue, options.expectedJsonType, false)) {
            throw new Error("JsonTsMapper | class : '" + options.className + "' | serialize : type missmatch for property " + options.classPropertyName + ": expected type : " + options.expectedJsonType.name + ", serialized type : " + typeof jsonValue);
        }
        return jsonValue;
    };
    /**
     * @param {?} jsonValue
     * @param {?=} expectedJsonType
     * @param {?=} isArray
     * @return {?}
     */
    JsonTsMapperService.prototype.checkType = /**
     * @param {?} jsonValue
     * @param {?=} expectedJsonType
     * @param {?=} isArray
     * @return {?}
     */
    function (jsonValue, expectedJsonType, isArray) {
        if (expectedJsonType === void 0) { expectedJsonType = Any; }
        if (isArray === void 0) { isArray = false; }
        if (isArray) {
            if (Array.isArray(jsonValue)) {
                if (jsonValue.length) {
                    jsonValue = jsonValue[0];
                }
                else {
                    return true;
                }
            }
            else {
                return false;
            }
        }
        switch (expectedJsonType) {
            case Any:
                return true;
            case String:
                return typeof jsonValue === 'string';
            case Number:
                return typeof jsonValue === 'number';
            case Boolean:
                return typeof jsonValue === 'boolean';
            default:
                return typeof jsonValue === 'object';
        }
    };
    JsonTsMapperService.decorators = [
        { type: Injectable, args: [{
                    providedIn: 'root'
                },] }
    ];
    /** @nocollapse */ JsonTsMapperService.ngInjectableDef = defineInjectable({ factory: function JsonTsMapperService_Factory() { return new JsonTsMapperService(); }, token: JsonTsMapperService, providedIn: "root" });
    return JsonTsMapperService;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @abstract
 * @template JsonType, TsType
 */
var  /**
 * @abstract
 * @template JsonType, TsType
 */
AbstractJsonConverter = /** @class */ (function () {
    function AbstractJsonConverter() {
    }
    return AbstractJsonConverter;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { JsonObject, JsonObjectOptions, JsonProperty, Optional, NotNull, CustomConverter, JsonConverter, Any, JsonTsMapperService, AbstractJsonConverter };

//# sourceMappingURL=thorolf-json-ts-mapper.js.map