import 'reflect-metadata';
import { Context } from './json-converter';
export declare class JsonTsMapperService {
    /**
     * Serialize objct to 'json object'
     * @param  obj object or array to serialize
     * @return    json object
     */
    serialize<T>(instance: T | T[], context?: Context): any;
    /**
     * Serialize objct to json string
     * @param  obj object or array to serialize
     * @return    json string
     */
    serializeToString<T>(instance: T | T[], context?: Context): string;
    /**
     * Deserialize an instance or array of {T}
     * @param obj object, array or json string to deserialize
     * @param clazz class to map object
     * @return instance or array of {T}
     */
    deserialize<T>(obj: object | object[] | string, clazz: new () => T, context?: Context): T | T[];
    /**
     * Deserialize {T}
     * @param obj object or json string to deserialize
     * @param clazz class to map object
     * @return instance of {T}
     */
    deserializeObject<T>(obj: object | string, clazz: new () => T, context?: Context): T;
    /**
     * Deserialize array of {T}
     * @param obj array or json string to deserialize
     * @param clazz class to map objects
     * @return T[]
     */
    deserializeArray<T>(obj: object[] | string, clazz: new () => T, context?: Context): T[];
    private deserialize_loopProperties;
    private deserializePropertyArray;
    private deserializeProperty;
    private serialize_loopProperties;
    private serializePropertyArray;
    private serializeProperty;
    checkType(jsonValue: any, expectedJsonType?: new () => {}, isArray?: boolean): boolean;
}
